/*
 * Copyright (c) 2011 Wolfgang Solfrank.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the  
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Wolfgang Solfrank.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <MiXen.h>
#include <atomic.h>
#include <memory.h>
#include <string.h>
#include <systm.h>

#include <machine/pte.h>

extern void _start(uint64_t, start_info_t *);
extern int _data;

#define	STACKSIZE	(2 * PAGESIZE)

/*
 * Setup the mapping expected by MiXen:
 *
 * At the top of the lower half of the virtual address space are
 * the page tables.
 *
 * Other than that, we use addresses starting at 0x2000000000 upwards.
 * At the start of this we reserve some space for virtual addresses,
 * where we map some other memory (like e.g. foreign pages), i.e.
 * during initialization we allocate page table pages for these, but
 * don't map any memory here.
 *
 * Above that we map all of the available memory (except of course,
 * the memory used for our code, data, bss, stack, the start_info,
 * console and xenstore pages and the memory needed for page tables.)
 *
 * The upper half of the address space contains the hypervisor, the
 * MiXen kernel within the top 2 GB (as mandated by -mcmodel=kernel).
 * Just below the kernel, we map any virtual addresses we need for
 * malloc'ed memory.
 *
 * Note that currently we never unmap any page table pages required
 * at any time.  Note also, that for now, we simply allocate virtual
 * address space top down without ever recovering it.
 *
 * Note things not (or incorrectly documented in <xen.h>:
 * 1.	Virtual address 0 (actually VIRT_BASE) to the start of the
 *	kernel image is mapped with the pseudo frames starting at 0.
 * 2.	The machine address given by start_info->shared_info is
 *	not mapped.
 * 3.	The machine frames start_info->store_mfn and
 *	start_info->console.domU.mfn are mapped between the start_info
 *	and the start of the page tables.
 * 4.	Before changing to a new L4 page table page, the new page
 *	must be pinned (via MMUEXT_PIN_L4_TABLE).  This is NOT
 *	true for any other page table pages and in fact isn't done
 *	for those on initial startup.
 * 5.	While the L4 page table page does always contain the mapping
 *	of Xen (i.e. has entries at indizes 0x200 and beyond), they
 *	don't have to be established (may not be?) in a new L4 ptp
 *	before switching to it.
 */

/* Top of the lower half of the address space. */
#define	TOP_LOWMEM	(VIRT_SPACE >> 1)

/* The memory needed to map some amount of memory. */
#define	PT_NEEDED(amt)	(((amt) >> PGSHIFT) * sizeof(pte_t))

/* We place the L1 page tables at the top of low memory. */
#define	L1_PTAB		(TOP_LOWMEM - PT_NEEDED(VIRT_SPACE))
#define	L1_PTE(addr)	(((pte_t *)L1_PTAB)[PG_LVLIDX(1, addr)])

/* The remaining page tables are just below that. */
#define	L2_PTAB		(L1_PTAB - PT_NEEDED(PT_NEEDED(VIRT_SPACE)))
#define	L2_PTE(addr)	(((pte_t *)L2_PTAB)[PG_LVLIDX(2, addr)])
#define	L3_PTAB		(L2_PTAB - PT_NEEDED(PT_NEEDED(PT_NEEDED(VIRT_SPACE))))
#define	L3_PTE(addr)	(((pte_t *)L3_PTAB)[PG_LVLIDX(3, addr)])
#define	L4_PTAB		(L3_PTAB - PAGESIZE)
#define	L4_PTE(addr)	(((pte_t *)L4_PTAB)[PG_LVLIDX(4, addr)])

static pte_t *const ptabs[] = {
    (pte_t *)L1_PTAB,
    (pte_t *)L2_PTAB,
    (pte_t *)L3_PTAB,
    (pte_t *)L4_PTAB
};

#define	pfn2va(pfn)	((void *)(((uint64_t)(pfn) << PGSHIFT) + VIRT_BASE))
#define	va2pfn(va)	(((uint64_t)(va) - VIRT_BASE) >> PGSHIFT)
#define	mfn2pa(mfn)	((uint64_t)(mfn) << PGSHIFT)
#define	pa2mfn(pa)	((uint64_t)(pa) >> PGSHIFT)

/* Provide readonly access to the page directory. */
static uint64_t cr3;

/*
 * Handle available memory via a chained list of available pages.
 * For now, we only allow to allocate exactly 1 page at a time
 * (by enlarging smaller requests and panicing on larger ones.)
 */
static struct page {
    struct page *next;
} *freepages;
uint64_t freecnt;

/* Base of memory for our own allocations (virtual & physical): */
#define	MALLOCMEM	((void *)(VIRT_SPACE / 8))

uint64_t virtfree[(VIRTMAP + sizeof(uint64_t) * 8 - 1) / sizeof(uint64_t) / 8];

#ifdef	RAMDISK
/* Base address of ramdisk: */
void *ramdisk = (void *)((char *)MALLOCMEM
			  + (VIRTMAP + RSRV_PAGES) * PAGESIZE * MEM_STRIDE);
uint64_t disksize;
#endif

/*
 * Initially freemem points to the top of mapped memory, where we
 * steal pages top down during initialization.
 */
static char *freemem;

/* MMU update command batcher. */
#define	MMUCCNT	64
static mmu_update_t mmucmds[MMUCCNT];
static int mmuccnt;

/*
 * Flush mmu update queue.
 */
static void
mmucflush(void)
{
    mmu_update_t *mp;
    unsigned int done;

    for (mp = mmucmds; mmuccnt > 0; mp += done) {
	if (hypercall_mmu_update(mmucmds, mmuccnt, &done, DOMID_SELF) < 0)
	    panic("mmu_update");
	mmuccnt -= done;
    }
}

/*
 * Enter some mmu update command into the queue.
 */
static void
mmucmd(uint64_t maddr, pte_t val)
{
    maddr &= ~3;	/* SHOULDN'T THERE BE A #define FOR THIS? XXX */
    mmucmds[mmuccnt].ptr = maddr | MMU_NORMAL_PT_UPDATE;
    mmucmds[mmuccnt].val = val;
    if (++mmuccnt >= MMUCCNT)
	mmucflush();
}

/* MMU extension command batcher. */
#define	MMUECNT	16
static mmuext_op_t mmuexts[MMUCCNT];
static int mmuecnt;

/*
 * Flush mmu extension queue.
 */
static void
mmueflush(void)
{
    mmuext_op_t *mp;
    unsigned int done;

    mmucflush();
    for (mp = mmuexts; mmuecnt > 0; mp += done) {
	if (hypercall_mmuext_op(mp, mmuecnt, &done, DOMID_SELF) < 0)
	    panic("mmuext");
	mmuecnt -= done;
    }
}

/*
 * Enter some mmu extended command into the queue.
 */
static void
mmuext(int cmd, uint64_t arg1, uint64_t arg2)
{
    mmuexts[mmuecnt].cmd = cmd;
    mmuexts[mmuecnt].arg1.linear_addr = arg1;
    mmuexts[mmuecnt].arg2.src_mfn = arg2;
    if (++mmuecnt >= MMUECNT)
	mmueflush();
}

/*
 * During construction of the page table directory, we first write virtual
 * addresses into the page table pages under construction and translate
 * them later.
 */
static void
mapvirt(pte_t *l4tab, void *vaddr, void *page, pte_t rwx)
{
    struct {
	char *v;
	void *p;
    } todo[10];
    int cnt, lvl;
    uint64_t idx;
    pte_t *ptab;

    todo[0].v = vaddr;
    todo[0].p = page;
    for (cnt = 1; --cnt >= 0;) {
	vaddr = todo[cnt].v;
	page = todo[cnt].p;
	ptab = l4tab;
	for (lvl = PGLVLS;; lvl--) {
	    idx = PG_LVLIDX(lvl, vaddr) & PGLVLIDX;
	    if (lvl == 1)
		break;
	    if (ptab[idx] & PTE_PRSNT)
		ptab = (pte_t *)UN_MASK_VADDR(ptab[idx] & PTE_PADDR);
	    else {
		ASSERT(cnt < __arraycount(todo));
		freemem -= PAGESIZE;
		ASSERT((uint64_t)freemem >= start_info->pt_base + start_info->nr_pt_frames * PAGESIZE);
		memset(freemem, 0, PAGESIZE);
		ptab[idx] = ((pte_t)freemem & PTE_PADDR) | PTE_PRSNT | PTE_RW;
		ptab = (pte_t *)freemem;
		todo[cnt].v = (char *)ptabs[lvl - 2] \
				+ (PG_LVLIDX(lvl, vaddr) << PGSHIFT);
		todo[cnt++].p = ptab;
	    }
	}
	ptab[idx] = ((pte_t)page & PTE_PADDR) | PTE_PRSNT | rwx;
	rwx = PTE_NX;	/* Page table pages are readonly and no-exec. */
    }
}

/*
 * Given a physical address of a page table page from the startup mapping,
 * return its virtual address.
 */
static uint64_t *
o_vaddr(uint64_t pa)
{
    uint64_t *mfnl = (uint64_t *)start_info->mfn_list;
    uint64_t cnt, pfn;

    pa = pa2mfn(pa);
    pfn = va2pfn(start_info->pt_base);
    for (cnt = start_info->nr_pt_frames; cnt > 0; cnt--, pfn++)
	if (pa == mfnl[pfn])
	    return pfn2va(pfn);
    panic("o_vaddr: not in pt_base");
}

/*
 * Given a virtual address within the startup mapping,
 * return the physical address of its pte.
 */
static uint64_t
o_pte(void *vaddr)
{
    pte_t *ptab = (pte_t *)start_info->pt_base;
    uint64_t idx, pa;
    int lvl;

    for (lvl = PGLVLS;; lvl--) {
	idx = PG_LVLIDX(lvl, vaddr) & PGLVLIDX;
	ASSERT(ptab[idx] & PTE_PRSNT);
	if (lvl == 1)
	    return pa + idx * sizeof *ptab;
	pa = ptab[idx] & PTE_PADDR;
	ptab = o_vaddr(pa);
    }
}

/*
 * Iterate over the (not yet installed) l4tab and convert the
 * virtual addresses to physical ones, and mark them readonly
 * in the old page tables while here.
 */
static void
virt2phys(pte_t *l4tab)
{
    uint64_t *mfnl = (uint64_t *)start_info->mfn_list;
    int i1, i2, i3, i4;
    pte_t *l3tab, *l2tab, *l1tab;
    uint64_t pg;

    for (i4 = PGLVLIDX; i4 >= 0; i4--)
	if (l4tab[i4] & PTE_PRSNT) {
	    l3tab = (pte_t *)UN_MASK_VADDR(l4tab[i4] & PTE_PADDR);
	    l4tab[i4] &= ~PTE_PADDR;
	    l4tab[i4] |= mfn2pa(mfnl[va2pfn(l3tab)]);
	    for (i3 = PGLVLIDX; i3 >= 0; i3--)
		if (l3tab[i3] & PTE_PRSNT) {
		    l2tab = (pte_t *)UN_MASK_VADDR(l3tab[i3] & PTE_PADDR);
		    l3tab[i3] &= ~PTE_PADDR;
		    l3tab[i3] |= mfn2pa(mfnl[va2pfn(l2tab)]);
		    for (i2 = PGLVLIDX; i2 >= 0; i2--)
			if (l2tab[i2] & PTE_PRSNT) {
			    l1tab = (pte_t *)
				UN_MASK_VADDR(l2tab[i2] & PTE_PADDR);
			    l2tab[i2] &= ~PTE_PADDR;
			    l2tab[i2] |= mfn2pa(mfnl[va2pfn(l1tab)]);
			    for (i1 = PGLVLIDX; i1 >= 0; i1--)
				if (l1tab[i1] & PTE_PRSNT) {
				    pg = UN_MASK_VADDR(l1tab[i1] & PTE_PADDR);
				    l1tab[i1] &= ~PTE_PADDR;
				    l1tab[i1] |= mfn2pa(mfnl[va2pfn(pg)]);
				}
			    mmucmd(o_pte(l1tab), l2tab[i2] & ~PTE_RW);
			}
		    mmucmd(o_pte(l2tab), l3tab[i3] & ~PTE_RW);
		}
	    mmucmd(o_pte(l3tab), l4tab[i4] & ~PTE_RW);
	}
    pg = mfn2pa(mfnl[va2pfn(l4tab)]);
    mmucmd(o_pte(l4tab), pg | PTE_PRSNT | PTE_NX);
}

/*
 * Try to establish a mapping of a page at the specified virtual address.
 * Returns the level where the page was mapped (0 if there was already
 * some mapping at vaddr).
 */
static int
mappage(void *vaddr, uint64_t paddr, pte_t rwx, int domap)
{
    uint64_t ppte = L1_PTE(L4_PTAB) & PTE_PADDR;
    pte_t pte;
    int lvl;

    ASSERT((paddr & ~PTE_PADDR) == 0);

    for (lvl = PGLVLS; --lvl >= 0;) {
	pte = ptabs[lvl][PG_LVLIDX(lvl + 1, vaddr)];
	if (!(pte & PTE_PRSNT)) {
	    if (lvl > 0) {
		/*
		 * We need to establish a new page table page.
		 */
		if (mappage(&ptabs[lvl - 1][PG_LVLIDX(lvl, vaddr)],
			    paddr, PTE_NX, 1) != 1) {
		    /*
		     * This page was needed for further recursive mapping,
		     * so there still isn't a mapping for the required pte.
		     */
		    return lvl + 1;
		}
		/*
		 * Since we map this page as a page table page,
		 * generalize the mapping for the lower layers to read-write,
		 * but no-exec, as our code was already handled.
		 * The level 1 mapping may further restrict access rights.
		 */
		rwx = PTE_NX | PTE_RW;
		mmuext(MMUEXT_CLEAR_PAGE, pa2mfn(paddr), 0);
		mmueflush();
	    }
	    if (lvl > 0 || domap)
		mmucmd(ppte + (PG_LVLIDX(lvl + 1, vaddr) & PGLVLIDX)
				* sizeof(pte_t),
			paddr | PTE_PRSNT | rwx);
	    if (lvl > 0)
		mmucflush();
	    return lvl + 1;
	}
	ppte = pte & PTE_PADDR;
    }
    return 0;
}

/*
 * We are given the physical address of a new page of available memory.
 * We try to map it at increasing addresses, starting at MALLOCMEM.
 * However, if the neccessary page table pages to map the page
 * are not yet available, we use the given page for this mapping first.
 * Note that we map only every other virtual address in order to leave
 * a guard page between pages!
 */
static void
freepage(uint64_t paddr)
{
    static char *addr = (char *)MALLOCMEM;
    static uint64_t *mfnp;
    struct page *pg;

    for (;;) {
	if (addr < (char *)MALLOCMEM + VIRTMAP * PAGESIZE * MEM_STRIDE) {
	    if (mappage(addr, paddr, PTE_NX | PTE_RW, 0) == 1) {
		addr += PAGESIZE * MEM_STRIDE;
		continue;
	    }
	    return;
	}

#ifdef	RAMDISK
	/*
	 * If we already found enough memory for RSRV_PAGES,
	 * map the rest of memory as ramdisk, i.e. be sure to
	 * map the initial ramdisk at this address.
	 */
	if (addr >= (char *)ramdisk) {
	    ASSERT(addr < (char *)L4_PTAB);

	    if ((addr - (char *)ramdisk) < start_info->mod_len * MEM_STRIDE) {
		if (!mfnp)
		    mfnp = (uint64_t *)start_info->mfn_list
				+ va2pfn(start_info->mod_start);

		if (mappage(addr, paddr, PTE_NX | PTE_RW, 0) == 1) {
		    ASSERT(mappage(addr, mfn2pa(*mfnp++), PTE_RW | PTE_NX, 1)
				== 1);
		    addr += PAGESIZE * MEM_STRIDE;
		    disksize += PAGESIZE / DEV_BSIZE;
		    continue;
		}
		return;
	    }
	}
	/* Fall through to map remaining ram as end of ramdisk. */
#endif
	if (mappage(addr, paddr, PTE_NX | PTE_RW, 1) == 1) {
#ifdef	RAMDISK
	    if (addr >= (char *)ramdisk)
		disksize += PAGESIZE / DEV_BSIZE;
	    else
#endif
	    {
		mmucflush();
		pg = (struct page *)addr;
		pg->next = freepages;
		freepages = pg;
		freecnt++;
	    }
	    addr += PAGESIZE * MEM_STRIDE;
	}
	return;
    }
}

/*
 * For now we allocate memory a page at a time.
 */
void *
malloc(size_t size)
{
    struct page *pg;

    while (1) {
	pg = freepages;
	ASSERT(pg);
	if (atomic_cas_ptr(&freepages, pg, pg->next) == pg) {
	    atomic_dec_64(&freecnt);
	    return pg;
	}
    }
}

void
free(void *p)
{
    struct page *pg = p;

    while (1) {
	pg->next = freepages;
	if (atomic_cas_ptr(&freepages, pg->next, pg) == pg->next) {
	    atomic_inc_64(&freecnt);
	    break;
	}
    }
}

#ifdef	RAMDISK
uint64_t
vaddr2mfn(void *p)
{
    return pa2mfn(L1_PTE(p) & PTE_PADDR);
}
#endif

/*
 * Map the given reference somewhere and return its virtual address.
 */
void *
mapforeign(domid_t dom, grant_ref_t ref, grant_handle_t *handle, uint64_t *mfn)
{
    gnttab_map_grant_ref_t gmap;
    int idx, bit;
    uint64_t bits;
    void *pg;

    while (1) {
	for (idx = 0; idx < __arraycount(virtfree); idx++)
	    if ((bits = virtfree[idx]) != (uint64_t)-1L)
		break;
	ASSERT(idx < __arraycount(virtfree));
	for (bit = 0; bit < sizeof virtfree[0] * 8; bit++)
	    if (!(bits & (1L << bit)))
		break;
	ASSERT(bit < sizeof virtfree[0] * 8);
	if (atomic_cas_64(&virtfree[idx], bits, bits | (1 << bit)) == bits)
	    break;
    }
    idx = idx * sizeof virtfree[0] * 8 + bit;
    ASSERT(idx < VIRTMAP);
    pg = (char *)MALLOCMEM + idx * MEM_STRIDE * PAGESIZE;

    gmap.host_addr = (uint64_t)pg;
    gmap.flags = GNTMAP_host_map;
    gmap.ref = ref;
    gmap.dom = dom;

    ASSERT(hypercall_grant_table_op(GNTTABOP_map_grant_ref, &gmap, 1) == 0);
    ASSERT(gmap.status == GNTST_okay);

    *handle = gmap.handle;
    *mfn = gmap.dev_bus_addr;
    return pg;
}

void
unmapforeign(grant_handle_t handle, void *p, uint64_t mfn)
{
    gnttab_unmap_grant_ref_t gunmap;

    gunmap.host_addr = (uint64_t)p;
    gunmap.dev_bus_addr = mfn;
    gunmap.handle = handle;

    ASSERT(hypercall_grant_table_op(GNTTABOP_unmap_grant_ref, &gunmap, 1) == 0);
    ASSERT(gunmap.status == GNTST_okay);
}

void
setup_mem(void)
{
    uint64_t tom;
    void *stack;

    /*
     * Determine the end of the initially mapped memory.
     */
    tom = (uint64_t)&tom;
    tom = (tom + SOD_ALIGNMENT - 1) & ~(SOD_ALIGNMENT - 1);
    if (tom - (uint64_t)&tom < SOD_GAP)
	tom += SOD_ALIGNMENT;

    /*
     * Switch to a new, larger stack.
     */
    stack = (void *)tom;
    switchstack(stack);
    freemem = stack - STACKSIZE;
}

void
init_mem(void)
{
    uint64_t *mfnl = (uint64_t *)start_info->mfn_list;
    uint64_t pfn, lpfn;
    pte_t *l4tab;
    char *p, *stack = freemem;

    /*
     * Use the next lower page for the new L4 page directory.
     */
    freemem -= PAGESIZE;
    l4tab = (void *)freemem;
    memset(l4tab, 0, PAGESIZE);
    mapvirt(l4tab, (void *)L4_PTAB, l4tab, PTE_NX);

    /*
     * Map any mapped memory up to the start of the original page tables.
     * If there was an initial ram disk, exclude that, it will be mapped later.
     * Thus we map the kernel image, the mfn_list and the start_info,
     * xenstore and console pages.
     */
    for (p = (char *)&_start; p < (char *)&_data; p += PAGESIZE)
	mapvirt(l4tab, p, p, 0);
    while (p < (char *)start_info->pt_base) {
#ifdef	RAMDISK
	if ((uint64_t)p == start_info->mod_start) {
	    p += (start_info->mod_len + PAGESIZE - 1) & ~(PAGESIZE - 1);
	    continue;
	}
#endif
	mapvirt(l4tab, p, p, PTE_NX | PTE_RW);
	p += PAGESIZE;
    }

    /*
     * Verify that xenstore & console page follow the start_info page.
     */
    pfn = va2pfn(start_info);
    ASSERT(mfnl[pfn + 1] == start_info->store_mfn);
    ASSERT(mfnl[pfn + 2] == start_info->console.domU.mfn);

    /*
     * Map the new stack.
     */
    for (p = stack; p < stack + STACKSIZE; p += PAGESIZE)
	mapvirt(l4tab, p, p, PTE_NX | PTE_RW);

    /*
     * Convert the virtual addresses in the constructed page tables
     * to physical addresses (marking them read only on the fly).
     */
    virt2phys(l4tab);

    /*
     * Install the new page tables.
     */
    cr3 = mfn2pa(mfnl[va2pfn(l4tab)]);
    mmuext(MMUEXT_PIN_L4_TABLE, mfnl[va2pfn(l4tab)], 0);
    mmuext(MMUEXT_NEW_BASEPTR, mfnl[va2pfn(l4tab)], 0);
    mmuext(MMUEXT_UNPIN_TABLE, mfnl[va2pfn(start_info->pt_base)], 0);
    mmueflush();

    /*
     * Map the shared_info page 2 pages below the stack.
     * Note that due to this address, we know that the mapping
     * doesn't need any additional page table pages.
     */
    stack -= 2 * PAGESIZE;
    ASSERT(freemem < stack);
    mappage(stack, start_info->shared_info, PTE_NX | PTE_RW, 1);
    mmucflush();
    shared_info = (struct shared_info *)stack;

    /*
     * Map any other memory at MALLOCMEM.
#ifdef	RAMDISK
     * Ramdisk mapping is taken char of by freepage as well (see above).
#endif
     * The first VIRTMAP pages don't get real memory
     * only enough page table pages to allow mapping things here.
     *
     * Start mapping the memory from the old page tables up to
     * the start of memory stolen for the mapping till now.
     */
    lpfn = va2pfn(freemem);
    for (pfn = va2pfn(start_info->pt_base); pfn < lpfn; pfn++)
	freepage(mfn2pa(mfnl[pfn]));

    for (pfn = va2pfn(stack + 2 * PAGESIZE + STACKSIZE);
	 pfn < start_info->nr_pages; pfn++)
	freepage(mfn2pa(mfnl[pfn]));

    mmucflush();

    printf("%ld kB memory available\n", freecnt * PAGESIZE / 1024);
#ifdef	RAMDISK
    printf("Emulating %lu sector ram disk\n", disksize);
#endif
}

uint64_t
rdcr3(void)
{
    return cr3;
}
