/*
 * Copyright (c) 2011 Wolfgang Solfrank.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the  
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Wolfgang Solfrank.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Example pseudo driver implementation of a ramdisk for MiXen.
 */
#include <MiXen.h>
#include <atomic.h>
#include <memory.h>
#include <string.h>
#include <systm.h>

#include <event_channel.h>
#include <grant_table.h>

#include <io/blkif.h>

static struct guest {
    void		*xs_arg;
    blkif_sring_t	*s_ring;
    blkif_back_ring_t	b_ring;
    grant_handle_t	handle;
    uint64_t		mfn;
    domid_t		dom;
    evtchn_port_t	event;
    int			proto32;
} guests[NGUESTS];

/*
 * Signal the other side that we handled some stuff.
 */
static void
be_signal(struct guest *g)
{
    evtchn_send_t send;

    xen_wmb();
    send.port = g->event;
    ASSERT(hypercall_event_channel_op(EVTCHNOP_send, &send) == 0);
}

/*
 * Enter into gc a copy operation from our virtual address sp to
 * offset off in page ref in domain dom with length len.
 */
static void
gcread(gnttab_copy_t *gc,
       grant_ref_t ref, uint16_t off, domid_t dom,
       char *sp, uint16_t len)
{
    gc->source.u.gmfn = vaddr2mfn(sp);
    gc->source.domid = DOMID_SELF;
    gc->source.offset = (uint64_t)sp & (PAGESIZE - 1);
    gc->dest.u.ref = ref;
    gc->dest.domid = dom;
    gc->dest.offset = off;
    gc->len = len;
    gc->flags = GNTCOPY_dest_gref;
}

/*
 * Enter into gc a copy operation from offset off in page ref in domain dom
 * to our virtual address dp with length len.
 */
static void
gcwrite(gnttab_copy_t *gc,
       grant_ref_t ref, uint16_t off, domid_t dom,
       char *dp, uint16_t len)
{
    gc->source.u.ref = ref;
    gc->source.domid = dom;
    gc->source.offset = off;
    gc->dest.u.gmfn = vaddr2mfn(dp);
    gc->dest.domid = DOMID_SELF;
    gc->dest.offset = (uint64_t)dp & (PAGESIZE - 1);
    gc->len = len;
    gc->flags = GNTCOPY_source_gref;
}

/*
 * Enter response into ringbuffer.
 */
static void
respond(blkif_response_t *rsp, uint64_t id, uint8_t op, int16_t status)
{
    rsp->id = id;
    rsp->operation = op;
    rsp->status = status;
}

/*
 * Protocol conversion.
 * These structures are only used within the convert_* routines.
 */
struct blkif_seg_32 {
    grant_ref_t	gref;
    uint8_t	first_sect;
    uint8_t	last_sect;
    uint16_t	pad;
} __attribute__((packed));
struct blkif_req_32 {
    uint8_t	operation;
    uint8_t	nr_segments;
    uint16_t	handle;
    uint64_t	id;
    blkif_sector_t sector_number;
    struct blkif_seg_32	seg[BLKIF_MAX_SEGMENTS_PER_REQUEST];
} __attribute__((packed));
struct blkif_rsp_32 {
    uint64_t	id;
    uint8_t	operation;
    uint16_t	status;
} __attribute__((packed));
struct blkif_sring_32 {
    RING_IDX	req_prod;
    RING_IDX	req_event;
    RING_IDX	rsp_prod;
    RING_IDX	rsp_event;
    uint8_t	pad[48];
    union {
	struct blkif_req_32	req;
	struct blkif_rsp_32	rsp;
    } ring[1];
} __attribute__((packed));
struct blkif_ring_32 {
    RING_IDX rsp_prod_pvt;
    RING_IDX req_cons;
    unsigned int nr_ents;
    struct blkif_sring_32 *sring;
};
/*
 * Get some x86_32-abi request out of some ring buffer at index i.
 */
static void
convert_request(blkif_request_t *req, void *p, RING_IDX i)
{
    struct blkif_ring_32 *ring = p;
    struct blkif_req_32 *sp;

    sp = RING_GET_REQUEST(ring, i);
    req->operation = sp->operation;
    req->nr_segments = sp->nr_segments;
    req->handle = sp->handle;
    req->id = sp->id;
    req->sector_number = sp->sector_number;
    memcpy(req->seg, sp->seg, sizeof req->seg);
}

/*
 * Put our native response into an x86_32-abi response at dst, index i.
 */
static void
convert_response(void *p, RING_IDX i, blkif_response_t *rsp)
{
    struct blkif_ring_32 *ring = p;
    struct blkif_rsp_32 *dp;

    dp = RING_GET_RESPONSE(ring, i);
    dp->id = rsp->id;
    dp->operation = rsp->operation;
    dp->status = rsp->status;
}

/*
 * Handle some frontend request(s).
 */
static void
be_handler(void *arg)
{
    struct guest *g = arg;
    blkif_back_ring_t *ring = &g->b_ring;
    blkif_request_t req;
    blkif_response_t rsp;
    int work, notify;
    uint64_t sector, soff, doff;
    char *dp;
    size_t len, rem;
    gnttab_copy_t gcs[BLKIF_MAX_SEGMENTS_PER_REQUEST * 2];
    int gi, i;
    RING_IDX idx;
    static uint32_t ring_lock;

    while (1) {
	/*
	 * Is there work to be done?
	 */
	RING_FINAL_CHECK_FOR_REQUESTS(ring, work);
	if (!work)
	    break;

	idx = ring->req_cons;
	if (g->proto32)
	    convert_request(&req, ring, idx);
	else
	    memcpy(&req, RING_GET_REQUEST(ring, idx), sizeof req);

	if (atomic_cas_32(&ring->req_cons, idx, idx + 1) != idx)
	    continue;

	switch (req.operation) {

	case BLKIF_OP_READ:
#ifdef	DEBUG
	    DPRINTF(DBG_BACKEND, "reading sector %lu", req.sector_number);
	    gi = -1;
	    for (i = 0;
		 i < req.nr_segments && i < BLKIF_MAX_SEGMENTS_PER_REQUEST;
		 i++)
		gi += req.seg[i].last_sect - req.seg[i].first_sect + 1;
	    if (gi > 0)
		DPRINTF(DBG_BACKEND, " - %lu", req.sector_number + gi);
	    DPRINTF(DBG_BACKEND, "\n");
#endif

	    ASSERT(req.nr_segments > 0);
	    ASSERT(req.nr_segments <= BLKIF_MAX_SEGMENTS_PER_REQUEST);

	    sector = req.sector_number * XEN_BSIZE;
	    for (gi = i = 0; i < req.nr_segments; i++) {

		/* Find sector in ramdisk. */
		soff = sector / PAGESIZE;
		soff *= PAGESIZE * MEM_STRIDE;
		soff += sector % PAGESIZE;
		dp = (char *)ramdisk + soff;

		/* Find destination address. */
		doff = req.seg[i].first_sect * XEN_BSIZE;
		len = req.seg[i].last_sect - req.seg[i].first_sect + 1;
		len *= XEN_BSIZE;

		ASSERT(len <= PAGESIZE);
		/* Clip data at end of ramdisk. */
		if (soff + len > disksize * DEV_BSIZE * MEM_STRIDE)
		    len = disksize * DEV_BSIZE * MEM_STRIDE - soff;
		if (len == 0)
		    /* Should we be able to tell 'em about the short read? */
		    break;

		sector += len;
		rem = PAGESIZE - soff % PAGESIZE;
		if (len > rem) {
		    /*
		     * Split the transfer to not cross a page boundary.
		     */
		    ASSERT(gi < __arraycount(gcs) - 1);
		    gcread(&gcs[gi++],
			   req.seg[i].gref, doff, g->dom, dp, rem);
		    len -= rem;
		    dp += rem + (MEM_STRIDE - 1) * PAGESIZE;
		    doff += rem;
		}
		ASSERT(gi < __arraycount(gcs) - 1);
		gcread(&gcs[gi++], req.seg[i].gref, doff, g->dom, dp, len);
	    }

	    /*
	     * All set up, launch the copy.
	     */
	    ASSERT(hypercall_grant_table_op(GNTTABOP_copy, gcs, gi) == 0);
	    for (i = 0; i < gi; i++)
		ASSERT(gcs[i].status == GNTST_okay);

	    /* Tell 'em that we are done. */
	    respond(&rsp, req.id, req.operation, BLKIF_RSP_OKAY);
	    break;

	case BLKIF_OP_WRITE:
#ifdef	DEBUG
	    DPRINTF(DBG_BACKEND, "writing sector %lu", req.sector_number);
	    gi = -1;
	    for (i = 0;
		 i < req.nr_segments && i < BLKIF_MAX_SEGMENTS_PER_REQUEST;
		 i++)
		gi += req.seg[i].last_sect - req.seg[i].first_sect + 1;
	    if (gi > 0)
		DPRINTF(DBG_BACKEND, " - %lu", req.sector_number + gi);
	    DPRINTF(DBG_BACKEND, "\n");
#endif
	    ASSERT(req.nr_segments > 0);
	    ASSERT(req.nr_segments <= BLKIF_MAX_SEGMENTS_PER_REQUEST);

	    sector = req.sector_number * XEN_BSIZE;
	    for (gi = i = 0; i < req.nr_segments; i++) {

		/* Find source address. */
		soff = req.seg[i].first_sect * XEN_BSIZE;
		len = req.seg[i].last_sect - req.seg[i].first_sect + 1;
		len *= XEN_BSIZE;

		/* Find sector in ramdisk. */
		doff = sector / PAGESIZE;
		doff *= PAGESIZE * MEM_STRIDE;
		doff += sector % PAGESIZE;
		dp = (char *)ramdisk + doff;

		ASSERT(len <= PAGESIZE);
		/* Clip data at end of ramdisk. */
		if (doff + len > disksize * DEV_BSIZE * MEM_STRIDE)
		    len = disksize * DEV_BSIZE * MEM_STRIDE - doff;
		if (len == 0)
		    /* Tell 'em about the short write? */
		    break;

		sector += len;
		rem = PAGESIZE - doff % PAGESIZE;
		if (len > rem) {
		    /*
		     * Split the transfer to not cross a page boundary.
		     */
		    ASSERT(gi < __arraycount(gcs) - 1);
		    gcwrite(&gcs[gi++],
			   req.seg[i].gref, soff, g->dom, dp, rem);
		    len -= rem;
		    dp += rem + (MEM_STRIDE - 1) * PAGESIZE;
		    soff += rem;
		}
		ASSERT(gi < __arraycount(gcs) - 1);
		gcwrite(&gcs[gi++], req.seg[i].gref, soff, g->dom, dp, len);
	    }

	    /*
	     * All set up, launch the copy.
	     */
	    ASSERT(hypercall_grant_table_op(GNTTABOP_copy, gcs, gi) == 0);
	    for (i = 0; i < gi; i++)
		ASSERT(gcs[i].status == GNTST_okay);

	    /* Tell 'em that we are done. */
	    respond(&rsp, req.id, req.operation, BLKIF_RSP_OKAY);
	    break;

	default:
	    /* Tell 'em that we don't support this operation (yet?). */
	    DPRINTF(DBG_BACKEND, "unsupported operation %d\n", req.operation);
	    respond(&rsp, req.id, req.operation, BLKIF_RSP_EOPNOTSUPP);
	    break;
	}

	while (atomic_cas_32(&ring_lock, 0, 1) != 0)
	    yield();

	if (g->proto32)
	    convert_response(ring, ring->rsp_prod_pvt, &rsp);
	else
	    memcpy(RING_GET_RESPONSE(ring, ring->rsp_prod_pvt),
		&rsp, sizeof rsp);

	ring->rsp_prod_pvt++;

	RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(ring, notify);
	ring_lock = 0;

	if (notify)
	    be_signal(g);
    }
}

/*
 * Set parameters in xenstore for this device.
 * The argument is an opaque pointer to be given to xs_send.
 * Our return value is the argument to backend_setup and backend_destroy below.
 */
void *
backend_setparams(void *arg)
{
    int i, free;

    while (1) {
	free = -1;
	for (i = 0; i < __arraycount(guests); i++) {
	    if (guests[i].xs_arg == 0)
		free = i;
	    else if (guests[i].xs_arg == arg)
		break;
	}

	/* Slot already allocated for this guest. */
	if (i < __arraycount(guests))
	    break;

	/* No free slot for any guest. */
	if (free < 0)
	    return 0;

	/* Occupy the found free slot. */
	i = free;
	if (atomic_cas_ptr(&guests[i].xs_arg, 0, arg) == 0)
	    break;
    }
    guests[i].event = -1;

#ifdef	STDARG_FIXED
    xs_write(arg, "/sectors", "U", disksize * DEV_BSIZE / XEN_BSIZE);
    /* Should we take "/mode" into account here? */
    xs_write(arg, "/info", "u", 0);
    xs_write(arg, "/sector-size", "u", DEV_BSIZE);
#else
    xs_uint64(arg, "/sectors", disksize * DEV_BSIZE / XEN_BSIZE);
    /* Should we take "/mode" into account here? */
    xs_uint64(arg, "/info", 0);
    xs_uint64(arg, "/sector-size", DEV_BSIZE);
#endif

    return &guests[i];
}

/*
 * Ok, xenstore got us all information, start handling requests.
 */
void
backend_setup(void *arg, domid_t dom, grant_ref_t ring, evtchn_port_t event,
	int proto32)
{
    struct guest *g = arg;
    evtchn_bind_interdomain_t bind;

    if (g->event != -1) {
	/*
	 * We already saw these parameters.
	 * Coming here again is probably due to a transaction retry.
	 */
	ASSERT(g->dom == dom);
	return;
    }
    g->dom = dom;
    bind.remote_dom = dom;
    bind.remote_port = event;
    ASSERT(hypercall_event_channel_op(EVTCHNOP_bind_interdomain, &bind) == 0);
    g->event = bind.local_port;
    g->s_ring = mapforeign(dom, ring, &g->handle, &g->mfn);
    g->proto32 = proto32;
    BACK_RING_INIT(&g->b_ring, g->s_ring, PAGESIZE);
    install_handler(g->event, be_handler, arg);
}

/*
 * The frontend was torn down, free any associated resources.
 */
void
backend_destroy(void *arg)
{
    struct guest *g = arg;
    evtchn_close_t close;

    if (g->s_ring) {
	close.port = g->event;
	ASSERT(hypercall_event_channel_op(EVTCHNOP_close, &close) == 0);
	uninstall_handler(g->event);
	unmapforeign(g->handle, g->s_ring, g->mfn);
	g->s_ring = 0;
    }
    atomic_swap_ptr(&g->xs_arg, 0);
}
