/*
 * Copyright (c) 2011 Wolfgang Solfrank.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the  
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Wolfgang Solfrank.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <MiXen.h>
#include <atomic.h>
#include <memory.h>
#include <stdlib.h>
#include <string.h>
#include <systm.h>

#include <sched.h>
#include <vcpu.h>

struct start_info *start_info;
struct shared_info *shared_info;

#ifdef	DEBUG
int debug = 0;
#endif

static struct {
    evtchn_port_t port;
/* pollevents polls for all ports: */
#define	NOPORT0		((evtchn_port_t)-1)
/* pollevents polls for all ports except console and xenstore: */
#define	NOPORTn		((evtchn_port_t)-2)
    void (*handler)(void *);
    void *arg;
} evnthandler[NPORTS];

static void banner(void);
static void cmdline(char *);

int
main(struct start_info *sinfo)
{
    int i;
    start_info = sinfo;

    for (i = 0; i < NPORTS; i++)
	evnthandler[i].port = NOPORT0;
    cmdline(start_info->cmd_line);
    init_hypercall_page();
    setup_mem();
    init_console();
    banner();
    init_mem();

    /* No interrupts ever! */
    shared_info->vcpu_info[0].evtchn_upcall_mask = 1;

    init_xenstore();

    find_vcpus();

    while (1)
	pollevents(NOPORT0);
}

static void
banner(void)
{
    printf("\nCopyright (c) 2011\n");
    printf("    Wolfgang Solfrank.  All rights reserved.\n\n");
    printf("%s\n", version);
}

static void
cmdline(char *cmd)
{
#ifdef	DEBUG
    /*
     * For now we only allow 'debug=xxx' for debugging output
     * and ignore anything else.
     */
    if (!memcmp(cmd, "debug=", sizeof "debug=" - 1)) {
	debug = strtol(cmd + sizeof "debug=" - 1, 0, 16);
	DPRINTF(DBG_ALL, "debugging = 0x%x\n", debug);
    }
#endif
}

/*
 * For additional cpus we don't poll xenstore or console.
 */
void
run_vcpu(void)
{
    while (1)
	pollevents(NOPORTn);
}

void
halt(void)
{
#if 0	/* This will be enabled when we are ready. */
    /*
     * On automatic shutdown, xenconsole may not yet have completely
     * swallowed our output.
     * For now, we rely on the user to destroy the domain.
     */
    struct sched_shutdown shutdown;

    shutdown.reason = SHUTDOWN_poweroff;
    hypercall_sched_op(SCHEDOP_shutdown, &shutdown);
#else
    while (1)
	yield();
#endif
}

void
yield(void)
{
    hypercall_sched_op(SCHEDOP_yield, 0);
}

static int
ispending(evtchn_port_t port)
{
    size_t idx;
    unsigned long bit;

    if (shared_info == 0) {
	/* nothing we can do about it. */
	return 0;
    }

    ASSERT(port < sizeof shared_info->evtchn_pending * 8);
    idx = sizeof shared_info->evtchn_pending[0] * 8;
    bit = 1UL << (port % idx);
    idx = port / idx;
    if (shared_info->evtchn_pending[idx] & bit) {
	atomic_and_64(&shared_info->evtchn_pending[idx], ~bit);
	return 1;
    }
    return 0;
}

void
pollevents(evtchn_port_t port)
{
    sched_poll_t poll;
    evtchn_port_t pollports[NPORTS];
    int i, j;

    poll.nr_ports = 0;
    poll.timeout = 0;
    switch (port) {
    case NOPORT0:
	for (i = 0; i < NPORTS; i++)
	    if ((port = evnthandler[i].port) != NOPORT0) {
		pollports[poll.nr_ports] = port;
		poll.nr_ports++;
	    }
	poll.ports.p = pollports;
	break;
    case NOPORTn:
	for (i = 2; i < NPORTS; i++)
	    if ((port = evnthandler[i].port) != NOPORT0) {
		pollports[poll.nr_ports] = port;
		poll.nr_ports++;
	    }
	poll.ports.p = pollports;
	break;
    default:
	poll.ports.p = &port;
	poll.nr_ports = 1;
	break;
    }
    hypercall_sched_op(SCHEDOP_poll, &poll);
    for (i = 0; i < poll.nr_ports; i++)
	if (ispending(poll.ports.p[i]))
	    for (j = 0; j < NPORTS; j++)
		if (evnthandler[j].port == poll.ports.p[i]
		    && evnthandler[j].handler) {
		    (*evnthandler[j].handler)(evnthandler[j].arg);
		    break;
		}
}

void
install_handler(evtchn_port_t port, void (*handler)(void *), void *arg)
{
    int i;

    for (i = 0; i < NPORTS; i++)
	if (atomic_cas_32(&evnthandler[i].port, (uint32_t)-1, (uint32_t)port)
	    == (uint32_t)-1) {
	    evnthandler[i].arg = arg;
	    evnthandler[i].handler = handler;
	    return;
	}
    panic("no slot for handler");
}

void
uninstall_handler(evtchn_port_t port)
{
    int i;

    for (i = 0; i < NPORTS; i++)
	if (atomic_cas_32(&evnthandler[i].port, (uint32_t)port, (uint32_t)-1)
	    == (uint32_t)port) {
	    evnthandler[i].handler = 0;
	    break;
	}
}
