/*
 * Copyright (c) 2011 Wolfgang Solfrank.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the  
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Wolfgang Solfrank.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdarg.h>
#include <string.h>
#include <MiXen.h>

static int
fmtuint(void (*out)(void *, const char *, size_t), void *arg,
    int base, unsigned long i)
{
    static const char digit[] = "0123456789abcdef";
    char buf[32], *cp = buf;
    unsigned long div;

    if (i == 0) {
	(*out)(arg, "0", 1);
	return 1;
    }
    for (div = 1; i / div >= base; div *= base);
    for (; div > 0; div /= base) {
	*cp++ = digit[i / div];
	i %= div;
    }
    (*out)(arg, buf, cp - buf);
    return cp - buf;
}

static size_t
vfprintf(void (*out)(void *, const char *, size_t),
	void *arg, const char *fmt, va_list ap)
{
    size_t len = 0, l;
    const char *cp, *cp1;
    char *s, c;
    int lng;
    long d;

    len = 0;
    for (cp = fmt; *cp; cp++) {
	switch (*cp) {
	default:
	    break;
	case '%':
	    l = cp - fmt;
	    (*out)(arg, fmt, l);
	    len += l;

	    lng = 0;
	    cp1 = cp;		/* cp1 -> '%' */

	    while (1) {
		switch (*++cp) {
		default:
		    /*
		     * Unknown conversion, just drop the '%'. (leave it alone?)
		     */
		    fmt = cp1 + 1;
		    cp = cp1;
		    break;
		case 'l':
		    lng = 1;
		    continue;
		case 's':
		    s = va_arg(ap, char *);
		    l = strlen(s);
		    (*out)(arg, s, l);
		    len += l;
		    fmt = cp + 1;
		    break;
		case 'c':
		    c = va_arg(ap, int);
		    (*out)(arg, &c, 1);
		    len++;
		    fmt = cp + 1;
		    break;
		case 'd':
		case 'u':
		    if (lng)
			d = va_arg(ap, long);
		    else
			d = va_arg(ap, int);
		    if (*fmt == 'd' && d < 0) {
			(*out)(arg, "-", 1);
			len++;
			d = -d;
		    }
		    len += fmtuint(out, arg, 10, d);
		    fmt = cp + 1;
		    break;
		case 'p':
		    lng = 1;
		case 'x':
		    if (lng)
			d = va_arg(ap, long);
		    else
			d = va_arg(ap, int);
		    len += fmtuint(out, arg, 16, d);
		    fmt = cp + 1;
		    break;
		}
		break;
	    }
	}
    }

    l = cp - fmt;
    (*out)(arg, fmt, l);
    len += l;

    return len;
}

size_t
vprintf(const char *fmt, va_list ap)
{
    return vfprintf(&console_write, 0, fmt, ap);
}

size_t
printf(const char *fmt, ...)
{
    va_list ap;
    size_t len;

    va_start(ap, fmt);
    len = vprintf(fmt, ap);
    va_end(ap);
    return len;
}

struct sbuf {
    char *buf;
    size_t avail;
};

static void
MemCpy(void *arg, const char *src, size_t len)
{
    struct sbuf *p = arg;

    if (len > p->avail - 1)
	len = p->avail - 1;
    memcpy(p->buf, src, len);
    p->buf += len;
    p->avail -= len;
    *p->buf = 0;
}

size_t
vsnprintf(char *buf, size_t len, const char *fmt, va_list ap)
{
    struct sbuf sb;

    sb.buf = buf;
    sb.avail = len;
    return vfprintf(&MemCpy, &sb, fmt, ap);
}

size_t
snprintf(char *buf, size_t len, const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    len = vsnprintf(buf, len, fmt, ap);
    va_end(ap);
    return len;
}
